@@ -1,13 +1,16 @@
-use super::canonical::IsCanonicalFn;
+use super::canonical::{CanonicalFnLock, IsCanonicalFn};
+use crate::curve::{Curve, CurveInfoLock, CurveLock, PrivateKey, PublicKey};
 use crate::hasher::{Digest, HashingAlgorithm, DIGEST_LEN};
+use crate::hd_node::{HDNODE_PRIVKEY_LEN, HDNODE_PUBKEY_LEN};
+use crate::signature::{RecoverableSignature, Signature, SIG_LEN};
 use core::marker::PhantomData;
 use core::{fmt, mem, ops};
+use std::os::raw::c_char;
 use std::sync::{Mutex, MutexGuard};
 
 pub const ECDSA_PUBKEY_COMPRESSED_LEN: usize = 33;
 pub const ECDSA_PUBKEY_UNCOMPRESSED_LEN: usize = 65;
 pub const ECDSA_PRIVKEY_LEN: usize = 32;
-pub const ECDSA_SIG_LEN: usize = 64;
 
 lazy_static! {
     static ref ECDSA_CURVE_LOCK: Mutex<()> = Mutex::new(());
@@ -19,12 +22,22 @@ pub struct EcdsaCurveLock {
     _lock: MutexGuard<'static, ()>,
 }
 
+impl CurveLock for EcdsaCurveLock {}
+
 impl EcdsaCurveLock {
-    unsafe fn as_ptr(&self) -> *const sys::ecdsa_curve {
+    pub(crate) unsafe fn as_ptr(&self) -> *const sys::ecdsa_curve {
         self.curve
     }
 }
 
+impl CurveInfoLock for EcdsaCurveInfoLock {
+    type CurveLock = EcdsaCurveLock;
+    #[inline]
+    unsafe fn curve(&self) -> &EcdsaCurveLock {
+        &self.lock
+    }
+}
+
 impl ops::Deref for EcdsaCurveLock {
     type Target = sys::ecdsa_curve;
     fn deref(&self) -> &Self::Target {
@@ -32,6 +45,19 @@ impl ops::Deref for EcdsaCurveLock {
     }
 }
 
+#[doc(hidden)]
+pub struct EcdsaCurveInfoLock {
+    info: &'static sys::curve_info,
+    lock: EcdsaCurveLock,
+}
+
+impl ops::Deref for EcdsaCurveInfoLock {
+    type Target = sys::curve_info;
+    fn deref(&self) -> &Self::Target {
+        self.info
+    }
+}
+
 pub trait EcdsaCurve {
     #[doc(hidden)]
     unsafe fn curve_lock() -> EcdsaCurveLock;
@@ -91,19 +117,18 @@ trait EcdsaCurveExt: EcdsaCurve {
         priv_key: &[u8; ECDSA_PRIVKEY_LEN],
         digest: &[u8; DIGEST_LEN],
         is_canonical: Option<IsCanonicalFn>,
-    ) -> Option<([u8; ECDSA_SIG_LEN], u8)> {
-        let mut sig = [0; ECDSA_SIG_LEN];
+    ) -> Option<([u8; SIG_LEN], u8)> {
+        let mut sig = [0; SIG_LEN];
         let mut by = 0;
         let res = unsafe {
             let curve = Self::curve_lock();
-            curve.set_is_canonical_func(is_canonical);
             sys::ecdsa_sign_digest(
                 curve.as_ptr(),
                 priv_key.as_ptr(),
                 digest.as_ptr(),
                 sig.as_mut_ptr(),
                 &mut by,
-                Some(EcdsaCurveLock::is_canonical),
+                curve.is_canonical_fn(is_canonical),
             )
         };
         if res == 0 {
@@ -117,12 +142,11 @@ trait EcdsaCurveExt: EcdsaCurve {
         hasher_type: sys::HasherType,
         data: &[u8],
         is_canonical: Option<IsCanonicalFn>,
-    ) -> Option<([u8; ECDSA_SIG_LEN], u8)> {
-        let mut sig = [0; ECDSA_SIG_LEN];
+    ) -> Option<([u8; SIG_LEN], u8)> {
+        let mut sig = [0; SIG_LEN];
         let mut by = 0;
         let res = unsafe {
             let curve = Self::curve_lock();
-            curve.set_is_canonical_func(is_canonical);
             sys::ecdsa_sign(
                 curve.as_ptr(),
                 hasher_type,
@@ -131,7 +155,7 @@ trait EcdsaCurveExt: EcdsaCurve {
                 data.len() as u32,
                 sig.as_mut_ptr(),
                 &mut by,
-                Some(EcdsaCurveLock::is_canonical),
+                curve.is_canonical_fn(is_canonical),
             )
         };
         if res == 0 {
@@ -142,7 +166,7 @@ trait EcdsaCurveExt: EcdsaCurve {
     }
     fn verify_digest(
         pub_key: &[u8; ECDSA_PUBKEY_COMPRESSED_LEN],
-        sig: &[u8; ECDSA_SIG_LEN],
+        sig: &[u8; SIG_LEN],
         digest: &[u8; DIGEST_LEN],
     ) -> bool {
         let res = unsafe {
@@ -159,7 +183,7 @@ trait EcdsaCurveExt: EcdsaCurve {
     fn verify(
         pub_key: &[u8; ECDSA_PUBKEY_COMPRESSED_LEN],
         hasher_type: sys::HasherType,
-        sig: &[u8; ECDSA_SIG_LEN],
+        sig: &[u8; SIG_LEN],
         data: &[u8],
     ) -> bool {
         let res = unsafe {
@@ -176,7 +200,7 @@ trait EcdsaCurveExt: EcdsaCurve {
         res == 0
     }
     fn recover_pub_from_sig(
-        sig: &[u8; ECDSA_SIG_LEN],
+        sig: &[u8; SIG_LEN],
         digest: &[u8; DIGEST_LEN],
         recid: u8,
     ) -> Option<[u8; ECDSA_PUBKEY_UNCOMPRESSED_LEN]> {
@@ -202,22 +226,48 @@ trait EcdsaCurveExt: EcdsaCurve {
 impl<T> EcdsaCurveExt for T where T: EcdsaCurve {}
 
 macro_rules! ecdsa_curve {
-    ($name:ident, $ty:path) => {
+    ($name:ident, $curve:path, $info:path, $name_ptr:path) => {
         #[derive(Clone, Copy, Debug, PartialEq, Eq)]
         pub struct $name;
         impl EcdsaCurve for $name {
             #[inline]
             unsafe fn curve_lock() -> EcdsaCurveLock {
                 EcdsaCurveLock {
-                    curve: &$ty,
+                    curve: &$curve,
                     _lock: ECDSA_CURVE_LOCK.lock().unwrap(),
                 }
             }
         }
+        impl Curve for $name {
+            type PublicKey = EcdsaPublicKey<Self>;
+            type PrivateKey = EcdsaPrivateKey<Self>;
+            type CurveInfoLock = EcdsaCurveInfoLock;
+            #[inline]
+            unsafe fn curve_info_lock() -> EcdsaCurveInfoLock {
+                EcdsaCurveInfoLock {
+                    info: &$info,
+                    lock: Self::curve_lock(),
+                }
+            }
+            #[inline]
+            unsafe fn name_ptr() -> *const c_char {
+                $name_ptr.as_ptr()
+            }
+        }
     };
 }
-ecdsa_curve!(Secp256k1, sys::secp256k1);
-ecdsa_curve!(Nist256p1, sys::nist256p1);
+ecdsa_curve!(
+    Secp256k1,
+    sys::secp256k1,
+    sys::secp256k1_info,
+    sys::SECP256K1_NAME
+);
+ecdsa_curve!(
+    Nist256p1,
+    sys::nist256p1,
+    sys::nist256p1_info,
+    sys::NIST256P1_NAME
+);
 
 #[derive(Clone)]
 pub struct EcdsaPrivateKey<C: EcdsaCurve> {
@@ -256,7 +306,7 @@ impl<C: EcdsaCurve> EcdsaPrivateKey<C> {
         &self,
         data: D,
         is_canonical: Option<IsCanonicalFn>,
-    ) -> Option<RecoverableSignature> {
+    ) -> Option<RecoverableSignature<C>> {
         C::sign(&self.bytes, H::hasher_type(), data.as_ref(), is_canonical)
             .map(|(sig, by)| RecoverableSignature::new(Signature::from_bytes(sig), by))
     }
@@ -264,12 +314,23 @@ impl<C: EcdsaCurve> EcdsaPrivateKey<C> {
         &self,
         digest: &Digest,
         is_canonical: Option<IsCanonicalFn>,
-    ) -> Option<RecoverableSignature> {
+    ) -> Option<RecoverableSignature<C>> {
         C::sign_digest(&self.bytes, digest.as_bytes(), is_canonical)
             .map(|(sig, by)| RecoverableSignature::new(Signature::from_bytes(sig), by))
     }
 }
 
+impl<C: EcdsaCurve> PrivateKey for EcdsaPrivateKey<C> {
+    #[inline]
+    fn from_bytes_unchecked(bytes: [u8; HDNODE_PRIVKEY_LEN]) -> Self {
+        Self::from_bytes(bytes)
+    }
+    #[inline]
+    fn to_bytes(self) -> [u8; HDNODE_PRIVKEY_LEN] {
+        self.bytes
+    }
+}
+
 #[derive(Clone, PartialEq, Eq)]
 pub struct EcdsaPublicKey<C: EcdsaCurve> {
     bytes: [u8; ECDSA_PUBKEY_COMPRESSED_LEN],
@@ -302,12 +363,12 @@ impl<C: EcdsaCurve> EcdsaPublicKey<C> {
     pub fn serialize_uncompressed(&self) -> [u8; ECDSA_PUBKEY_UNCOMPRESSED_LEN] {
         C::uncompress_public_key(&self.bytes).unwrap()
     }
-    pub fn verify_digest(&self, signature: &Signature, digest: &Digest) -> bool {
+    pub fn verify_digest(&self, signature: &Signature<C>, digest: &Digest) -> bool {
         C::verify_digest(&self.bytes, &signature.serialize(), digest.as_bytes())
     }
     pub fn verify<H: HashingAlgorithm, D: AsRef<[u8]>>(
         &self,
-        signature: &Signature,
+        signature: &Signature<C>,
         data: D,
     ) -> bool {
         C::verify(
@@ -319,6 +380,17 @@ impl<C: EcdsaCurve> EcdsaPublicKey<C> {
     }
 }
 
+impl<C: EcdsaCurve> PublicKey for EcdsaPublicKey<C> {
+    #[inline]
+    fn from_bytes_unchecked(bytes: [u8; HDNODE_PUBKEY_LEN]) -> Self {
+        unsafe { Self::from_bytes_unchecked(bytes) }
+    }
+    #[inline]
+    fn to_bytes(self) -> [u8; HDNODE_PUBKEY_LEN] {
+        self.bytes
+    }
+}
+
 impl<C: EcdsaCurve> fmt::Debug for EcdsaPublicKey<C> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("EcdsaPublicKey")
@@ -328,28 +400,7 @@ impl<C: EcdsaCurve> fmt::Debug for EcdsaPublicKey<C> {
     }
 }
 
-#[derive(Clone, Debug, PartialEq, Eq)]
-pub struct Signature {
-    bytes: [u8; ECDSA_SIG_LEN],
-}
-
-impl Signature {
-    #[inline]
-    pub fn from_bytes(bytes: [u8; ECDSA_SIG_LEN]) -> Self {
-        Self { bytes }
-    }
-    pub fn from_slice(slice: &[u8]) -> Option<Self> {
-        if slice.len() == ECDSA_SIG_LEN {
-            let mut bytes = [0; ECDSA_SIG_LEN];
-            bytes.copy_from_slice(slice);
-            Some(Self::from_bytes(bytes))
-        } else {
-            None
-        }
-    }
-    pub fn serialize(&self) -> &[u8; ECDSA_SIG_LEN] {
-        &self.bytes
-    }
+impl<C: EcdsaCurve> Signature<C> {
     pub fn serialize_der(&self) -> Vec<u8> {
         let mut out = Vec::with_capacity(1 << 7);
         unsafe {
@@ -359,7 +410,7 @@ impl Signature {
         out
     }
     pub fn from_der(der: &[u8]) -> Option<Self> {
-        let mut out = [0; ECDSA_SIG_LEN];
+        let mut out = [0; SIG_LEN];
         let res =
             unsafe { sys::ecdsa_sig_from_der(der.as_ptr(), der.len() as u64, out.as_mut_ptr()) };
         if res == 0 {
@@ -370,53 +421,36 @@ impl Signature {
     }
 }
 
-#[derive(Clone, Debug, PartialEq, Eq)]
-pub struct RecoverableSignature {
-    signature: Signature,
-    recovery_byte: u8,
-}
-
-impl RecoverableSignature {
-    #[inline]
-    pub fn new(signature: Signature, recovery_byte: u8) -> Self {
-        Self {
-            signature,
-            recovery_byte,
-        }
-    }
-    pub fn recovery_byte(&self) -> u8 {
-        self.recovery_byte
-    }
-    pub fn recover_public_key<C: EcdsaCurve>(&self, digest: &Digest) -> Option<EcdsaPublicKey<C>> {
+impl<C: EcdsaCurve> RecoverableSignature<C> {
+    pub fn recover_public_key(&self, digest: &Digest) -> Option<EcdsaPublicKey<C>> {
         C::recover_pub_from_sig(
-            self.signature.serialize(),
+            self.signature().serialize(),
             digest.as_bytes(),
-            self.recovery_byte,
+            self.recovery_byte(),
         )
         .and_then(|bytes| EcdsaPublicKey::from_slice(&bytes))
     }
 }
 
-impl ops::Deref for RecoverableSignature {
-    type Target = Signature;
-    fn deref(&self) -> &Signature {
-        &self.signature
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::hasher::*;
 
+    #[test]
+    fn curve_name() {
+        assert_eq!("secp256k1", Secp256k1::name());
+        assert_eq!("nist256p1", Nist256p1::name());
+    }
+
     fn public_key_test_vector<C: EcdsaCurve>(priv_key_hex: &str, x_hex: &str, y_hex: &str) {
         let priv_key =
             EcdsaPrivateKey::<C>::from_slice(&hex::decode(priv_key_hex).unwrap()).unwrap();
         let pub_key = priv_key.public_key();
         assert!(pub_key.is_valid());
         let pub_key = pub_key.serialize_uncompressed();
-        assert_eq!(hex::encode(&pub_key[1..33]), x_hex.to_lowercase());
-        assert_eq!(hex::encode(&pub_key[33..]), y_hex.to_lowercase());
+        assert_eq!(pub_key[1..33], hex::decode(x_hex).unwrap());
+        assert_eq!(pub_key[33..], hex::decode(y_hex).unwrap());
     }
 
     #[test]
@@ -464,23 +498,6 @@ mod tests {
         );
     }
 
-    #[test]
-    fn secp256k1_multi_thread() {
-        let mut children = Vec::new();
-        for _ in 0..10 {
-            children.push(std::thread::spawn(|| {
-                public_key_test_vector::<Secp256k1>(
-                    "AA5E28D6A97A2479A65527F7290311A3624D4CC0FA1578598EE3C2613BF99522",
-                    "34F9460F0E4F08393D192B3C5133A6BA099AA0AD9FD54EBCCFACDFA239FF49C6",
-                    "0B71EA9BD730FD8923F6D25A7A91E7DD7728A960686CB5A901BB419E0F2CA232",
-                );
-            }))
-        }
-        for child in children {
-            child.join().unwrap();
-        }
-    }
-
     #[test]
     fn nist256p1_public_key_test_vectors() {
         public_key_test_vector::<Nist256p1>(
@@ -501,22 +518,22 @@ mod tests {
         let digest = digest::<Sha2, _>(message.as_ref());
         let signature = priv_key.sign_digest(&digest, None).unwrap();
         let sig = signature.serialize();
-        assert_eq!(hex::encode(&sig[..32]), r_hex.to_lowercase());
-        assert_eq!(hex::encode(&sig[32..]), s_hex.to_lowercase());
+        assert_eq!(sig[..32], hex::decode(r_hex).unwrap());
+        assert_eq!(sig[32..], hex::decode(s_hex).unwrap());
 
         let signature2 = priv_key.sign::<Sha2, _>(message.as_ref(), None).unwrap();
         assert_eq!(signature.serialize(), signature2.serialize());
 
         let der = signature.serialize_der();
         assert!(der.len() > 0);
-        let signature3 = Signature::from_der(&der).unwrap();
+        let signature3 = Signature::<C>::from_der(&der).unwrap();
         assert_eq!(signature.serialize(), signature3.serialize());
 
         let pub_key = priv_key.public_key();
         assert!(pub_key.verify::<Sha2, _>(&signature, message));
         assert!(pub_key.verify_digest(&signature, &digest));
 
-        let pub_key2 = signature.recover_public_key::<C>(&digest).unwrap();
+        let pub_key2 = signature.recover_public_key(&digest).unwrap();
         assert_eq!(pub_key.serialize(), pub_key2.serialize());
     }
 
@@ -534,7 +551,7 @@ mod tests {
         static ref COUNTER: Mutex<u8> = Mutex::new(0);
     }
 
-    fn test_is_canonical(_sig: &RecoverableSignature) -> bool {
+    fn test_is_canonical<C: EcdsaCurve>(_sig: RecoverableSignature<C>) -> bool {
         let mut counter = COUNTER.lock().unwrap();
         *counter += 1;
         if *counter < 5 {
@@ -556,4 +573,19 @@ mod tests {
             .unwrap();
         assert_eq!(*COUNTER.lock().unwrap(), 5);
     }
+
+    #[test]
+    fn ecdsa_multi_thread() {
+        let mut children = Vec::new();
+        for _ in 0..10 {
+            children.push(std::thread::spawn(|| {
+                secp256k1_public_key_test_vectors();
+                nist256p1_public_key_test_vectors();
+                nist256p1_signature_test_vectors();
+            }))
+        }
+        for child in children {
+            child.join().unwrap();
+        }
+    }
 }
