@@ -2,5 +2,9 @@
 extern crate lazy_static;
 extern crate trezor_crypto_sys as sys;
 
+pub mod curve;
 pub mod ecdsa;
+pub mod ed25519;
 pub mod hasher;
+pub mod hd_node;
+pub mod signature;
